/*
 * canonf50
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "usbdevice_fs.h"

#define TAG "canonf50"

#ifdef __GNUC__
#define __unused __attribute__((unused))
#else
#define __unused /**/
#endif

#define USB_VENDOR_ID_CANON      0x04a9
#define USB_DEVICE_ID_CANON_F50  0x2610

#define CMDPKT_LEN   10

struct params {
	char *devname;
	char *serial_number;
};

struct dev {
	struct params *par;
	int devfd;
	int intr_active;
	struct usbdevfs_urb intr_urb;
	unsigned char intr_buf[64];
};

#define STAT_WATER 0xFFFF00
#define STAT_WARM      0x01

void Usage(void);

void misc_fun(struct dev *d, unsigned fnum);
int get_status(struct dev *d, unsigned int *pstat);
void print_buf(struct dev *d, const char *buf, int len);
void intr_submit(struct dev *d);
void urb_reap(struct dev *d);
void parse_params(struct params *p, char **argv);
void print_hex(const unsigned char *buf, int len);

int main(int argc __unused, char **argv)
{
	struct params par;
	struct dev d;
	unsigned int intf = 0;		/* Hardcode zero for now */
	unsigned int statb;
	int errcnt;
	int cnt;
	int rc;
	enum { BUF_SIZE = 32768 };
	char buf[BUF_SIZE];

	parse_params(&par, argv+1);

	if (par.devname == NULL) {
		fprintf(stderr, TAG ": searching for device is not here yet\n");
		exit(1);
	}

	d.par = &par;

	if ((d.devfd = open(par.devname, O_RDWR)) == -1) {
		fprintf(stderr, TAG ": Can't open \"%s\": %s\n",
		    par.devname, strerror(errno));
		exit(1);
	}

	/* XXX Verify interface descriptor to have the right endpoints */

	/*
	 * Regular operations claim interfaces automatically, but it's nicer
	 * to users when such operations fail ahead of time.
	 */
	if (ioctl(d.devfd, USBDEVFS_CLAIMINTERFACE, &intf) != 0) {
		fprintf(stderr, TAG
		    ": Can't claim interface %u on \"%s\": %s\n",
		    intf, par.devname, strerror(errno));
		exit(1);
	}

	/* XXX Pretty useless so far... */
	intr_submit(&d);
	urb_reap(&d);

	/*
	 * This often fails after the printing,
	 * but it allows the F1 and F6 to proceed.
	 */
	(void) get_status(&d, &statb);

	misc_fun(&d, 1);	/* De-idle - interrupt? */
	misc_fun(&d, 6);	/* De-idle - start printing? */

	/* The 1 and 6 above do not make interrupts, but heck let's do it... */
	urb_reap(&d);
	if (!d.intr_active)
		intr_submit(&d);

	cnt = 0;
	for (;;) {
		if (get_status(&d, &statb) == 0 && (statb & STAT_WARM) != 0)
			break;
		if (++cnt >= 10) {
			fprintf(stderr, "Status not getting ready\n");
			exit(1);
		}
		sleep(1);
	}

	// print_bfuf(&d, "Hello, World!\r\n\014\033@",
	//         strlen("Hello, World!\r\n\014\033@"));

	errcnt = 0;
	for (;;) {
		rc = read(0, buf, BUF_SIZE);
		if (rc < 0) {
			fprintf(stderr, TAG ": Error reading input: %s\n",
			    strerror(errno));
			exit(1);
		}
		if (rc == 0)
			break;

		if (get_status(&d, &statb) == 0) {
			errcnt = 0;
			if ((statb&STAT_WATER) >= 160) {
				print_buf(&d, buf, rc);
			} else {
				sleep(1);
			}

			urb_reap(&d);
			if (!d.intr_active)
				intr_submit(&d);
		} else {
			if (++errcnt >= 10) {
				fprintf(stderr, TAG
				    ": Persistent status errors\n");
				exit(1);
			}
		}
	}

	misc_fun(&d, 7);	/* End of printing ? */

	cnt = 0;
	for (;;) {
		if (get_status(&d, &statb) == 0 && (statb & STAT_WARM) == 0)
			break;
		if (++cnt >= 90) {
			fprintf(stderr, "Status not getting ready\n");
			exit(1);
		}
		sleep(1);
	}

	misc_fun(&d, 2);	/* Set idling for interrupts ? */
	sleep(1);
	urb_reap(&d);

	return 0;
}

/*
 * This is the  fc 60 00 44 xx xx FN 00 00 00, low bits unset in lead bytes.
 */
void misc_fun(struct dev *d, unsigned fnum)
{
	unsigned char cmdpkt[CMDPKT_LEN];
	struct usbdevfs_bulktransfer bulkp;
	unsigned char reply[64];
	int rc;

	memset(cmdpkt, 0, CMDPKT_LEN);
	cmdpkt[0] = 0xfc;	// ?? Subfunction, 0x01 unset
	cmdpkt[1] = 0x60;	// ?? Subfunction
	cmdpkt[2] = 0x00;	// ?? Some kind of a bit, 0x01 unset
	cmdpkt[3] = 0x44;	// ?? Bits also
	cmdpkt[6] = fnum;

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x01;
	bulkp.len = CMDPKT_LEN;
	bulkp.data = cmdpkt;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Can't send F%u request to \"%s\": %d %s\n",
		    fnum, d->par->devname, errno, strerror(errno));
		exit(1);
	}
	if (rc < CMDPKT_LEN) {
		fprintf(stderr, TAG
		    ": F%u request transferred %d of %d\n",
		    fnum, rc, CMDPKT_LEN);
	}

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x82;
	bulkp.len = 64;
	bulkp.data = reply;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Error receiving F%u reply: %d %s\n",
		    fnum, errno, strerror(errno));
		return;
	}
	if (rc != 2) {
		fprintf(stderr, TAG ": reply size is not 2 (%d)\n", rc);
		return;
	}
	if (reply[0] != reply[1]) {
		fprintf(stderr, TAG
		    ": reply header bytes mismatch: 0x%02x 0x%02x\n",
		    reply[0], reply[1]);
		return;
	}
	printf("F%u reply %02x\n", fnum, reply[0]);
}

/*
 * Get Status
 */
int get_status(struct dev *d, unsigned int *pstat)
{
	unsigned char cmdpkt[CMDPKT_LEN];
	struct usbdevfs_bulktransfer bulkp;
	unsigned char reply[64];
	int i;
	unsigned char sum;
	int rc;

	memset(cmdpkt, 0, CMDPKT_LEN);
	cmdpkt[0] = 0xfd;	// ?? Subfunction   (0x01 reply or size)
	cmdpkt[1] = 0x60;	// ?? Subfunction
	cmdpkt[2] = 0x01;	// ?? Some kind of a bit (reply or size)
	cmdpkt[3] = 0x44;	// ?? Bits also
	cmdpkt[6] = 0;		// Looks like the function. 0 - request status.
	cmdpkt[7] = 0;		// Expected reply length MSB, without 2 hdr.
	cmdpkt[8] = 20;		// Expected reply length LSB, without 2 hdr.

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x01;
	bulkp.len = CMDPKT_LEN;
	bulkp.data = cmdpkt;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Can't send status request to \"%s\": %d %s\n",
		    d->par->devname, errno, strerror(errno));
		exit(1);
	}
	if (rc < CMDPKT_LEN) {
		fprintf(stderr, TAG
		    ": Status request transferred %d of %d\n",
		    rc, CMDPKT_LEN);
	}

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x82;
	bulkp.len = 64;
	bulkp.data = reply;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Error receiving status at \"%s\": %d %s\n",
		    d->par->devname, errno, strerror(errno));
		return -1;
	}
	if (rc != 22) {
		fprintf(stderr, TAG ": status size is not 2+20 (%d)\n", rc);
		return -1;
	}
	if (reply[0] != reply[1]) {
		fprintf(stderr, TAG
		    ": status header bytes mismatch: 0x%02x 0x%02x\n",
		    reply[0], reply[1]);
		return -1;
	}

	/*
	 * The "easy" way is to sum 20 bytes and compare to zero, but
	 * it is wrong. It would compare fine in case of all-zero buffer.
	 */
	sum = 0;
	for (i = 0; i < 19; i++) {
		sum += reply[2+i];
	}
	sum = -sum;
	if (sum != reply[2+19]) {
		fprintf(stderr, TAG ": Status checksum incorrect,"
		    " calculated 0x%x loaded 0x%x\n", sum, reply[2+19]);
	}

	printf("Status %02x  %02x %02x %02x\n", reply[0],
	     reply[7], reply[8], reply[10]);
	*pstat = (reply[7] << 16) | (reply[8] << 18) | reply[10];
	return 0;
}

void print_buf(struct dev *d, const char *buf, int len)
{
	int pad_len;
	unsigned char *cmdbuf;
	struct usbdevfs_bulktransfer bulkp;
	unsigned char reply[64];
	int rc;

	pad_len = (len + 3) & ~1;		/* Round up in a strange way */
	pad_len -= len;

	if ((cmdbuf = malloc(CMDPKT_LEN + len + pad_len)) == NULL) {
		fprintf(stderr, TAG ": No core\n");
		goto fail_none;
	}

	memset(cmdbuf, 0, CMDPKT_LEN);
	cmdbuf[0] = 0xfc;	// ?? Subfunction   (0x01 reply or size)
	cmdbuf[1] = 0x60;	// ?? Subfunction
	cmdbuf[2] = 0x00;	// ?? Some kind of a bit (reply or size)
	cmdbuf[3] = 0x44;	// ?? Bits also
	cmdbuf[4] = len >> 8;
	cmdbuf[5] = len & 0xFF;
	cmdbuf[6] = 5;		// Looks like the function. 5 - print.
	cmdbuf[7] = (len + pad_len) >> 8;
	cmdbuf[8] = (len + pad_len) & 0xFF;

	memcpy(cmdbuf + CMDPKT_LEN, buf, len);
	if (pad_len)
		memset(cmdbuf + CMDPKT_LEN + len, 0, pad_len);

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x01;
	bulkp.len = CMDPKT_LEN + len + pad_len;
	bulkp.data = cmdbuf;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Can't send print request to \"%s\": %d %s\n",
		    d->par->devname, errno, strerror(errno));
		exit(1);
	}
	if (rc < CMDPKT_LEN) {
		fprintf(stderr, TAG
		    ": Print request transferred %d of %d\n",
		    rc, CMDPKT_LEN);
	}

	memset(&bulkp, 0, sizeof(struct usbdevfs_bulktransfer));
	bulkp.ep = 0x82;
	bulkp.len = 64;
	bulkp.data = reply;
	bulkp.timeout = 1000;
	if ((rc = ioctl(d->devfd, USBDEVFS_BULK, &bulkp)) < 0) {
		fprintf(stderr, TAG
		    ": Error receiving print reply: %d %s\n",
		    errno, strerror(errno));
		exit(1);
	}
	if (rc != 2) {
		fprintf(stderr, TAG ": print reply size is not 2 (%d)\n", rc);
		goto fail;
	}
	if (reply[0] != reply[1]) {
		fprintf(stderr, TAG
		    ": print reply header bytes mismatch: 0x%02x 0x%02x\n",
		    reply[0], reply[1]);
		goto fail;
	}

	printf("Print reply %02x\n", reply[0]);
	free(cmdbuf);
	return;

fail:
	free(cmdbuf);
fail_none:
	return;
}

/*
 * Submit an interrupt URB.
 */
void intr_submit(struct dev *d)
{
	struct usbdevfs_urb *urb;
	int rc;

	if (d->intr_active) {
		fprintf(stderr, TAG ": IE intr submit\n");
		return;
	}

	urb = &d->intr_urb;
	memset(urb, 0, sizeof(struct usbdevfs_urb));
	urb->type = USBDEVFS_URB_TYPE_INTERRUPT;
	urb->endpoint = 0x83;	/* XXX Get from descriptors */
	urb->buffer = d->intr_buf;
	urb->buffer_length = 64;

	if ((rc = ioctl(d->devfd, USBDEVFS_SUBMITURB, urb)) < 0) {
		fprintf(stderr, TAG ": Can't submit intr URB (%d)\n", rc);
		return;
	}

	d->intr_active = 1;
}

/*
 * Reap all URBs (XXX or just one? What is better?)
 * This is asynchronous.
 */
void urb_reap(struct dev *d)
{
	struct usbdevfs_urb *urb_ret;
	int rc;

	if ((rc = ioctl(d->devfd, USBDEVFS_REAPURBNDELAY, &urb_ret)) < 0) {
		if (errno == EAGAIN) {
			printf("None reaped\n");
			return;
		}
		fprintf(stderr, TAG ": Can't reap intr URB (%d)\n", errno);
		return;
	}

	if (urb_ret != &d->intr_urb) {
		fprintf(stderr, TAG ": Wild URB reaped (0x%lx, expect 0x%lx)\n",
		    (long)urb_ret, (long)&d->intr_urb);
	} else {
		d->intr_active = 0;
	}
}

void parse_params(struct params *p, char **argv)
{
	char *arg;
	// long num;

	memset(p, 0, sizeof(struct params));

	while ((arg = *argv++) != NULL) {
		if (arg[0] == '-') {
			if (arg[1] == '\0')
				Usage();
			switch (arg[1]) {
			case 'd':
				if (arg[2] != '\0')
					Usage();
				if ((arg = *argv++) == NULL)
					Usage();
				p->devname = arg;
				break;
			case 'S':
				if (arg[2] != '\0')
					Usage();
				if ((arg = *argv++) == NULL)
					Usage();
				p->serial_number = arg;
				break;
			default:
				Usage();
			}
		} else {
			Usage();
		}
	}
}

void Usage()
{

	fprintf(stderr, "Usage: "
	    "canonf50 [-d /proc/bus/usb/001/002 ] [ -S 00SERNUM ]"
	    "\n");
	exit(1);
}

void print_hex(const unsigned char *buf, int len)
{
	int i;

	for (i = 0; i < len; i++) {
		if (i % 16 == 0) printf("%8x:", i);
		if (i % 2 == 0) printf(" ");
		printf("%02x", buf[i] & 0xFF);
		if (i % 16 == 15) printf("\n");
	}
	if (i % 16 != 15) printf("\n");
}
